"""
Print compilation information for the grav_sim C library.

Usage:
    python -m grav_sim [-p PATH]

This script searches for the compiled C library in the parent directory of the current file
(or a user-specified path), loads it using ctypes, and calls the `print_compilation_info`
function defined in the C library. Users can run this script to verify the installation
of the grav_sim library.
"""

import argparse
import ctypes
from pathlib import Path


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Print compilation information for the grav_sim C library."
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        default=None,
        help="Path to search for the grav_sim C library. Defaults to the parent directory of the current file.",
    )
    return parser.parse_args()


def main() -> None:
    args = parse_args()
    search_path = Path(args.path).parent if args.path else Path(__file__).parent.parent
    c_lib_files = [str(p) for p in search_path.rglob("*libgrav_sim*")]
    if len(c_lib_files) == 0:
        raise FileNotFoundError(f"C library not found from path: {search_path}")

    c_lib_path = c_lib_files[0]
    c_lib: ctypes.CDLL = ctypes.cdll.LoadLibrary(c_lib_path)
    c_lib.print_compilation_info()
    print(f"C library location: {c_lib_path}")


if __name__ == "__main__":
    main()



"""
Simulator module for grav_sim
"""

import ctypes
import time
import threading
from queue import Queue

import numpy as np

from .system import System, CosmologicalSystem
from .parameters import AccelerationParam, IntegratorParam, OutputParam, Settings


class Simulator:
    DAYS_PER_YEAR = 365.242189

    def __init__(self, c_lib: ctypes.CDLL) -> None:
        self.c_lib = c_lib

    def launch_simulation(
        self,
        system: System,
        acceleration_param: AccelerationParam,
        integrator_param: IntegratorParam,
        output_param: OutputParam,
        settings: Settings,
        tf: float,
    ) -> None:
        """Launch simulation

        Parameters
        ----------
        system : System
            Gravitational system
        acceleration_param : AccelerationParam
            Acceleration parameters
        integrator_param : IntegratorParam
            Integrator parameters
        output_param : OutputParam
            Output parameters
        settings : Settings
            Simulation settings
        tf : float
            Simulation time

        Raises
        -------
        RuntimeError
            If simulation fails
        """
        num_particles = ctypes.c_int32(system.num_particles)
        new_particle_ids = ctypes.POINTER(ctypes.c_int32)()
        new_x = ctypes.POINTER(ctypes.c_double)()
        new_v = ctypes.POINTER(ctypes.c_double)()
        new_m = ctypes.POINTER(ctypes.c_double)()

        is_exit = ctypes.c_bool(False)

        # Create a thread to stop the simulation if user sends KeyboardInterrupt
        def simulation_wrapper(c_lib_launch_simulation_python, return_queue, *args):
            return_queue.put(c_lib_launch_simulation_python(*args))

        queue: Queue = Queue()
        simulation_thread = threading.Thread(
            target=simulation_wrapper,
            args=(
                self.c_lib.launch_simulation_python,
                queue,
                ctypes.byref(num_particles),
                system.particle_ids.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)),
                system.x.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                system.v.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                system.m.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                ctypes.byref(new_particle_ids),
                ctypes.byref(new_x),
                ctypes.byref(new_v),
                ctypes.byref(new_m),
                ctypes.c_double(system.G),
                ctypes.c_int32(integrator_param._integrator),
                ctypes.c_double(integrator_param.dt),
                ctypes.c_double(integrator_param.tolerance),
                ctypes.c_double(integrator_param.initial_dt),
                ctypes.c_bool(integrator_param.whfast_remove_invalid_particles),
                ctypes.c_int32(acceleration_param._method),
                ctypes.c_double(acceleration_param.opening_angle),
                ctypes.c_double(acceleration_param.softening_length),
                ctypes.c_int32(acceleration_param.max_num_particles_per_leaf),
                ctypes.c_int32(output_param._method),
                output_param.output_dir.encode("utf-8"),
                ctypes.c_bool(output_param.output_initial),
                ctypes.c_double(output_param.output_interval),
                ctypes.c_int32(output_param._coordinate_output_dtype),
                ctypes.c_int32(output_param._velocity_output_dtype),
                ctypes.c_int32(output_param._mass_output_dtype),
                ctypes.c_int32(settings._verbose),
                ctypes.c_bool(settings.enable_progress_bar),
                ctypes.byref(is_exit),
                ctypes.c_double(tf),
            ),
        )

        ### Launch simulation ###
        try:
            simulation_thread.start()

            while simulation_thread.is_alive():
                time.sleep(0.05)

        except KeyboardInterrupt:
            print("\nKeyboardInterrupt detected. Stopping simulation...\n")
            is_exit.value = True
            simulation_thread.join()
            raise KeyboardInterrupt

        # End simulation and get return value
        simulation_thread.join()
        ret_val = queue.get()
        if ret_val != 0:
            raise RuntimeError(
                "Simulation failed. Please check the C library traceback."
            )

        # Update system with new data (just to be safe since
        # we don't know if the memory will be changed in the C library)
        system.particle_ids = np.ctypeslib.as_array(
            new_particle_ids,
            shape=(num_particles.value,),
        )
        system.x = np.ctypeslib.as_array(
            new_x,
            shape=(num_particles.value, 3),
        )
        system.v = np.ctypeslib.as_array(
            new_v,
            shape=(num_particles.value, 3),
        )
        system.m = np.ctypeslib.as_array(
            new_m,
            shape=(num_particles.value,),
        )
        system.num_particles = num_particles.value

    def launch_cosmological_simulation(
        self,
        system: CosmologicalSystem,
        output_param: OutputParam,
        settings: Settings,
        a_final: float,
        num_steps: int,
        pm_grid_size: int,
    ) -> None:
        """Launch simulation

        Parameters
        ----------
        system : CosmologicalSystem
            Cosmological system
        output_param : OutputParam
            Output parameters
        settings : Settings
            Simulation settings
        a_final : float
            Final scale factor
        num_steps : int
            Number of steps to simulate
        pm_grid_size : int
            Size of the Particle-Mesh grid

        Raises
        -------
        RuntimeError
            If simulation fails
        """
        num_particles = ctypes.c_int32(system.num_particles)
        new_particle_ids = ctypes.POINTER(ctypes.c_int32)()
        new_x = ctypes.POINTER(ctypes.c_double)()
        new_v = ctypes.POINTER(ctypes.c_double)()
        new_m = ctypes.POINTER(ctypes.c_double)()

        is_exit = ctypes.c_bool(False)

        # Create a thread to stop the simulation if user sends KeyboardInterrupt
        def simulation_wrapper(c_lib_launch_simulation_python, return_queue, *args):
            return_queue.put(c_lib_launch_simulation_python(*args))

        queue: Queue = Queue()
        simulation_thread = threading.Thread(
            target=simulation_wrapper,
            args=(
                self.c_lib.launch_cosmological_simulation_python,
                queue,
                ctypes.byref(num_particles),
                system.particle_ids.ctypes.data_as(ctypes.POINTER(ctypes.c_int32)),
                system.x.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                system.v.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                system.m.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                ctypes.byref(new_particle_ids),
                ctypes.byref(new_x),
                ctypes.byref(new_v),
                ctypes.byref(new_m),
                ctypes.c_double(system.h),
                ctypes.c_double(system.scale_factor),
                ctypes.c_double(system.omega_m),
                ctypes.c_double(system.omega_lambda),
                system.box_center.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                ctypes.c_double(system.box_width),
                ctypes.c_double(system.unit_mass_in_cgs),
                ctypes.c_double(system.unit_length_in_cgs),
                ctypes.c_double(system.unit_time_in_cgs),
                ctypes.c_int32(output_param._method),
                output_param.output_dir.encode("utf-8"),
                ctypes.c_bool(output_param.output_initial),
                ctypes.c_double(output_param.output_interval),
                ctypes.c_int32(output_param._coordinate_output_dtype),
                ctypes.c_int32(output_param._velocity_output_dtype),
                ctypes.c_int32(output_param._mass_output_dtype),
                ctypes.c_int32(settings._verbose),
                ctypes.c_bool(settings.enable_progress_bar),
                ctypes.byref(is_exit),
                ctypes.c_double(a_final),
                ctypes.c_int32(num_steps),
                ctypes.c_int32(pm_grid_size),
            ),
        )

        ### Launch simulation ###
        try:
            simulation_thread.start()

            while simulation_thread.is_alive():
                time.sleep(0.05)

        except KeyboardInterrupt:
            print("\nKeyboardInterrupt detected. Stopping simulation...\n")
            is_exit.value = True
            simulation_thread.join()
            raise KeyboardInterrupt

        # End simulation and get return value
        simulation_thread.join()
        ret_val = queue.get()
        if ret_val != 0:
            raise RuntimeError(
                "Simulation failed. Please check the C library traceback."
            )

        # Update system with new data (just to be safe since
        # we don't know if the memory will be changed in the C library)
        system.particle_ids = np.ctypeslib.as_array(
            new_particle_ids,
            shape=(num_particles.value,),
        )
        system.x = np.ctypeslib.as_array(
            new_x,
            shape=(num_particles.value, 3),
        )
        system.v = np.ctypeslib.as_array(
            new_v,
            shape=(num_particles.value, 3),
        )
        system.m = np.ctypeslib.as_array(
            new_m,
            shape=(num_particles.value,),
        )
        system.num_particles = num_particles.value



"""
This is a setup.py file for building the C extension using CMake.
It is modified from the pybind11 example:
https://github.com/pybind/cmake_example
"""

import os
import subprocess
import sys
from pathlib import Path

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext


# A CMakeExtension needs a sourcedir instead of a file list.
# The name must be the _single_ output extension from the CMake build.
# If you need multiple extensions, see scikit-build.
class CMakeExtension(Extension):
    def __init__(self, name: str, sourcedir: str = "") -> None:
        super().__init__(name, sources=[])
        self.sourcedir = os.fspath(Path(sourcedir).resolve())


class CMakeBuild(build_ext):
    def build_extension(self, ext: CMakeExtension) -> None:
        # Must be in this form due to bug in .resolve() only fixed in Python 3.10+
        ext_fullpath = Path.cwd() / self.get_ext_fullpath(ext.name)
        extdir = ext_fullpath.parent.resolve()

        # Using this requires trailing slash for auto-detection & inclusion of
        # auxiliary "native" libs

        debug = int(os.environ.get("DEBUG", 0)) if self.debug is None else self.debug
        cfg = "Debug" if debug else "Release"

        # CMake lets you override the generator - we need to check this.
        # Can be set with Conda-Build, for example.
        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

        # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
        # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
        # from Python.
        cmake_args = [
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}{os.sep}",
            f"-DPYTHON_EXECUTABLE={sys.executable}",
            f"-DCMAKE_BUILD_TYPE={cfg}",  # not used on MSVC, but no harm
        ]
        build_args = []
        # Adding CMake arguments set as environment variable
        # (needed e.g. to build for ARM OSx on conda-forge)
        if "CMAKE_ARGS" in os.environ:
            cmake_args += [item for item in os.environ["CMAKE_ARGS"].split(" ") if item]

        cmake_args += [f"-DVERSION_INFO={self.distribution.get_version()}"]

        build_temp = Path(self.build_temp) / ext.name
        if not build_temp.exists():
            build_temp.mkdir(parents=True)

        subprocess.run(
            ["cmake", ext.sourcedir, *cmake_args], cwd=build_temp, check=True
        )
        subprocess.run(
            ["cmake", "--build", ".", *build_args], cwd=build_temp, check=True
        )


# The information here can also be placed in setup.cfg - better separation of
# logic and declaration, and simpler if you include description/version in a file.
skip_cmake = os.environ.get("SKIP_CMAKE", "0") == "1"
if skip_cmake:
    ext_modules = []
else:
    ext_modules = [CMakeExtension("grav_sim", sourcedir="")]

setup(
    packages=["grav_sim"],
    ext_modules=ext_modules,
    cmdclass={"build_ext": CMakeBuild},
    package_data={"": ["src/*"]},
    include_package_data=True,
    zip_safe=False,
)



"""
Gravitational system
"""

import ctypes
import csv
from pathlib import Path
from typing import Optional

import numpy as np

from . import plotting
from . import utils


class System:
    """
    Class to represent a gravitational system.
    """

    # Built-in systems
    BUILT_IN_SYSTEMS = [
        "circular_binary_orbit",
        "eccentric_binary_orbit",
        "3d_helix",
        "sun_earth_moon",
        "figure-8",
        "pyth-3-body",
        "solar_system",
        "solar_system_plus",
    ]

    # Gravitational constant (AU^3 day^-2 M_sun^-1)
    G_DEFAULT = 0.000295912208284119496676630

    def __init__(self, c_lib: ctypes.CDLL) -> None:
        """Initialize System object"""
        self.c_lib = c_lib
        self.num_particles = 0
        self.particle_ids = np.zeros((0,), dtype=np.int32)
        self.x = np.zeros((0, 3), dtype=np.float64)
        self.v = np.zeros((0, 3), dtype=np.float64)
        self.m = np.zeros((0,), dtype=np.float64)
        self.G = self.G_DEFAULT

    def add(
        self,
        x: list | np.ndarray,
        v: list | np.ndarray,
        m: float | list | np.ndarray,
    ) -> None:
        """Add one or multiple particle to the system

        Parameters
        ----------
        x : list or np.ndarray
            Position vector(s) of the object(s)
        v : list or np.ndarray
            Velocity vector(s) of the object(s)
        m : float
            Mass(es) of the object(s)
        """
        self.x = np.vstack((self.x, np.array(x, dtype=np.float64)))
        self.v = np.vstack((self.v, np.array(v, dtype=np.float64)))
        self.m = np.hstack((self.m, m))

        if isinstance(m, (list, np.ndarray)):
            num_new_particles = len(m)
        else:
            num_new_particles = 1

        max_id = self.particle_ids.max() if len(self.particle_ids) > 0 else -1
        new_ids = np.arange(max_id + 1, max_id + 1 + num_new_particles, dtype=np.int32)
        self.particle_ids = np.hstack((self.particle_ids, new_ids))
        self.num_particles += num_new_particles

    def add_keplerian(
        self,
        semi_major_axis: float,
        eccentricity: float,
        inclination: float,
        argument_of_periapsis: float,
        longitude_of_ascending_node: float,
        true_anomaly: float,
        m: float,
        primary_particle_id: int,
    ):
        """
        Add a celestial body to the system using Keplerian elements

        Warning: This method use the G value from the system. Make sure
                 to set the correct G value before using this method.

        Parameters
        ----------
        semi_major_axis : float
            Semi-major axis
        eccentricity : float
            Eccentricity
        inclination : float
            Inclination
        argument_of_periapsis : float
            Argument of periapsis
        longitude_of_ascending_node : float
            Longitude of ascending node
        true_anomaly : float
            True anomaly
        m : float
            Mass
        primary_object_index : int
            Particle id of the primary object

        Raises
        ------
        ValueError
            If particle_id is not found in the system
        """

        primary_particle_idx = np.where(self.particle_ids == primary_particle_id)[0]
        if len(primary_particle_idx) == 0:
            raise ValueError(
                f"Particle ID {primary_particle_id} not found in the system"
            )
        primary_particle_idx = primary_particle_idx[0]

        x, v = utils.keplerian_to_cartesian(
            self.c_lib,
            semi_major_axis=semi_major_axis,
            eccentricity=eccentricity,
            inclination=inclination,
            argument_of_periapsis=argument_of_periapsis,
            longitude_of_ascending_node=longitude_of_ascending_node,
            true_anomaly=true_anomaly,
            total_mass=float(self.m[primary_particle_idx] + m),
            G=self.G,
        )
        self.add(x + self.x[primary_particle_idx], v + self.v[primary_particle_idx], m)

    def remove(
        self,
        particle_ids: int | list | np.ndarray,
    ) -> None:
        """Remove particle(s) from the system

        Parameters
        ----------
        particle_ids : int, list or np.ndarray
            particle_id(s) to remove

        Raises
        ------
        TypeError
            If indices is not an int, list or np.ndarray
        """
        if isinstance(particle_ids, int):
            particle_ids_arr = np.array([particle_ids])
        elif isinstance(particle_ids, list):
            particle_ids_arr = np.array(particle_ids)
        elif not isinstance(particle_ids, np.ndarray):
            raise TypeError("indices must be an int, list or np.ndarray")
        else:
            particle_ids_arr = particle_ids

        mask = np.isin(self.particle_ids, particle_ids_arr, invert=True)

        self.particle_ids = self.particle_ids[mask]
        self.x = self.x[mask]
        self.v = self.v[mask]
        self.m = self.m[mask]
        self.num_particles = len(self.particle_ids)

    def save(
        self,
        file_path: Optional[str | Path] = None,
    ) -> None:
        """Save system to a CSV file

        Parameters
        ----------
        file_path : str, optional
            File path to save the system, by default None

        Note
        ----
        This function has the following side effects:
        - Prints a message f"System \"{self.name}\" successfully saved to \"{file_path}\""
        """
        if file_path is None:
            file_path = Path(__file__).parent / "customized_systems.csv"

        with open(file_path, "w", newline="") as file:
            writer = csv.writer(file)

            # Write header
            writer.writerow(["particle_id", "m", "x", "y", "z", "vx", "vy", "vz"])

            # Write data
            for i in range(self.num_particles):
                writer.writerow(
                    [
                        self.particle_ids[i],
                        self.m[i],
                        self.x[i, 0],
                        self.x[i, 1],
                        self.x[i, 2],
                        self.v[i, 0],
                        self.v[i, 1],
                        self.v[i, 2],
                    ]
                )

        print(f'System successfully saved to "{file_path}"')

    @staticmethod
    def load_system(
        c_lib: ctypes.CDLL,
        file_path: str | Path,
    ) -> "System":
        """Load system from a CSV file

        Parameters
        ----------
        system_name : str
            Name of the system to load
        file_path : str
            File path to load the system from

        Raises
        ------
        FileNotFoundError
            If the file is not found
        """
        if file_path is not None:
            if isinstance(file_path, str):
                file_path = Path(file_path)

            if not file_path.is_file():
                raise FileNotFoundError(f"File not found: {file_path}")

        with open(file_path, "r") as file:
            reader = csv.reader(file)
            next(reader)  # Skip header

            system = System(c_lib)
            particle_ids = []
            for row in reader:
                system.add(
                    x=[float(row[2]), float(row[3]), float(row[4])],
                    v=[float(row[5]), float(row[6]), float(row[7])],
                    m=float(row[1]),
                )
                particle_ids.append(int(row[0]))
            system.particle_ids = np.array(particle_ids, dtype=np.int32)

        return system

    def center_of_mass_correction(self) -> None:
        """Set center of mass of position and velocity to zero"""
        M = np.sum(self.m)
        x_cm = np.einsum("i,ij->j", self.m, self.x) / M
        v_cm = np.einsum("i,ij->j", self.m, self.v) / M

        self.x -= x_cm
        self.v -= v_cm

    @staticmethod
    def get_built_in_system(c_lib: ctypes.CDLL, system_name: str) -> "System":
        """Get a built-in system from the C library

        Parameters
        ----------
        c_lib : ctypes.CDLL
            C dynamic-link library object
        system_name : str
            Name of the built-in system to be loaded.

        Returns
        -------
        System
            A System object with the loaded built-in system data.
        """
        # if system_name not in System.BUILT_IN_SYSTEMS:
        #     raise ValueError(
        #         f"Unknown system name: {system_name}. Available systems: {System.BUILT_IN_SYSTEMS}"
        #     )

        new_system = System(c_lib)

        num_particles = ctypes.c_int32()
        particle_ids = ctypes.POINTER(ctypes.c_int32)()
        x = ctypes.POINTER(ctypes.c_double)()
        v = ctypes.POINTER(ctypes.c_double)()
        m = ctypes.POINTER(ctypes.c_double)()
        G = ctypes.c_double()
        return_value = c_lib.load_built_in_system_python(
            system_name.encode("utf-8"),
            ctypes.byref(num_particles),
            ctypes.byref(particle_ids),
            ctypes.byref(x),
            ctypes.byref(v),
            ctypes.byref(m),
            ctypes.byref(G),
        )

        if return_value != 0:
            raise RuntimeError(
                f"Failed to load built-in system '{system_name}' from C library"
            )

        new_system.num_particles = num_particles.value
        new_system.G = G.value

        new_system.particle_ids = np.ctypeslib.as_array(
            particle_ids, shape=(num_particles.value,)
        ).copy()
        new_system.x = np.ctypeslib.as_array(x, shape=(num_particles.value, 3)).copy()
        new_system.v = np.ctypeslib.as_array(v, shape=(num_particles.value, 3)).copy()
        new_system.m = np.ctypeslib.as_array(m, shape=(num_particles.value,)).copy()

        c_lib.free_memory_int32(particle_ids)
        c_lib.free_memory_double(x)
        c_lib.free_memory_double(v)
        c_lib.free_memory_double(m)

        return new_system

    def plot_2d_system(
        self,
        colors: Optional[list[str]] = None,
        labels: Optional[list[str]] = None,
        legend: bool = False,
        xlabel: str = "$x$",
        ylabel: str = "$y$",
        title: Optional[str] = None,
        marker: str = "o",
        markersize: int = 6,
        save_fig: bool = False,
        save_fig_path: Optional[str | Path] = None,
    ) -> None:
        initial_state = np.concatenate(
            [
                self.x.flatten(),
                self.v.flatten(),
            ]
        )[np.newaxis, :]

        plotting.plot_2d_trajectory(
            initial_state,
            colors,
            labels,
            legend,
            xlabel,
            ylabel,
            title,
            marker,
            markersize,
            save_fig,
            save_fig_path,
        )

    def plot_3d_system(
        self,
        colors: Optional[list[str]] = None,
        labels: Optional[list[str]] = None,
        legend: bool = False,
        xlabel: str = "$x$",
        ylabel: str = "$y$",
        zlabel: str = "$z$",
        title: Optional[str] = None,
        marker: str = "o",
        markersize: int = 6,
        save_fig: bool = False,
        save_fig_path: Optional[str | Path] = None,
    ) -> None:
        initial_state = np.concatenate(
            [
                self.x.flatten(),
                self.v.flatten(),
            ]
        )[np.newaxis, :]

        plotting.plot_3d_trajectory(
            initial_state,
            colors,
            labels,
            legend,
            xlabel,
            ylabel,
            zlabel,
            title,
            marker,
            markersize,
            save_fig,
            save_fig_path,
        )


class CosmologicalSystem(System):
    """
    Class to represent a cosmological system.
    """

    def __init__(self, c_lib: ctypes.CDLL) -> None:
        super().__init__(c_lib)
        self.scale_factor = -1.0
        self.h = -1.0
        self.omega_m = -1.0
        self.omega_lambda = -1.0
        self.box_center = np.zeros((3,), dtype=np.float64)
        self.box_width = -1.0
        self.unit_mass_in_cgs = -1.0
        self.unit_length_in_cgs = -1.0
        self.unit_time_in_cgs = -1.0
